s/mf(mbox)/mf(mail-mbox)/g
s/mf(cmd)/mfp(OBSOLETE)/g
s/mf(deliver)/mf(mail-deliver)/g
s/mf(system)/mf(mail-system)/g
s/mf(tmp)/mf(mail-tmpdir)/g
s/mf(incorp)/mfp(OBSOLETE)/g
s/mf(mailchk)/mf(mail-interval)/g
s/mf(fdir)/mf(mail-directory)/g
s/mf(record)/mf(compose-fcc-folder)/g
s/mf(autoincorp)/mf(mail-auto-incorp)/g
s/mf(flush)/mf(mail-flush)/g
s/mf(recswap)/mf(compose-fcc-swap)/g
s/mf(leftscroll)/mf(disp-left-scroll)/g
s/mf(fdownbmp)/mf(viewer-bitmap-nomail)/g
s/mf(fupbmp)/mf(viewer-bitmap-mail)/g
s/mf(print)/mf(viewer-print)/g
s/mf(geom)/mf(viewer-geom)/g
s/mf(beep)/mf(viewer-beep-new)/g
s/mf(mesgstate)/mf(viewer-state)/g
s/mf(pipedir)/mf(viewer-pipe-dir)/g
s/mf(maxfold)/mf(menu-folders-max)/g
s/mf(quicksend)/mf(menu-quick-send)/g
s/mf(ignorefold)/mf(menu-folders-ignore)/g
s/mf(headlist-format)/mfp(CHANGED)/g
s/mf(reverse)/mf(headlist-reverse)/g
s/mf(headheight)/mf(headlist-height)/g
s/mf(format)/mfp(OBSOLETE)/g
s/mf(compgeom)/mf(compose-geom)/g
s/mf(showcc)/mf(compose-show-cc)/g
s/mf(savesend)/mf(compose-save-send)/g
s/mf(altedit)/mf(compose-alt-editor)/g
s/mf(autoaltedit)/mf(compose-alt-auto)/g
s/mf(alternates)/mf(compose-alternates)/g
s/mf(writebmp)/mf(compose-icon-bitmap)/g
s/mf(getfrom)/mfp(OBSOLETE)/g
s/mf(hdrstrip)/mf(header-strip)/g
s/mf(mime)/mf(header-mime)/g
s/mf(emacs)/mf(bind-emacs)/g
s/mf(prefix)/mf(insert-prefix)/g
s/mf(headers)/mf(insert-headers)/g
s/mf(headers-mime)/mfp(OBSOLETE)/g
s/mf(signature)/mf(insert-signature)/g
s/mf(encoder)/mf(insert-encoder)/g
s/mf(compress)/mf(insert-compress)/g
s/mf(compress_suffix)/mf(insert-compress-suffix)/g
s/mf(ispell)/mf(ispell-present)/g
s/set mf(tfont)/option add tkmail*Text.font/g
s/set mf(tcolor)/option add tkmail*Text.background/g
s/set mf(fcolor)/option add tkmail*Text.foreground/g
s/mf(mail-record)/mf(compose-fcc-folder)/g
s/mf(mail-record-swap)/mf(compose-fcc-swap)/g
s/mf(mail-record-forward)/mf(compose-fcc-forward)/g


